#ifndef EARTH_HPP
#define EARTH_HPP

#include "engine.hpp"
#include <vector>
using glm::vec2;
using glm::vec3;

vec3 lerp(vec3 A, vec3 B, float scalar) {
    return A + scalar*(B - A);
}

class Earth {
public:
    void initialize(Engine *engine, int slices, int stacks, float spherical);
    float isSpherical();
    void setSpherical(float spherical);
    vec3 getPosition(float latitude, float longitude);
    vec3 getNormal(float latitude, float longitude);
    void draw(bool textured);
    float getTargetSpherical();
    void setTargetSpherical(float s);
protected:
    int slices, stacks;
    int nVertices, nTriangles;
    float spherical, tarSpherical;
    Engine *engine;
    Texture texture;
    VertexBuffer vertBuffer, texCoordBuffer, normalBuffer;
    ElementBuffer faceBuffer;
    vec3* verts;
    vec3* normals;
    int* faces;
    vec2* texCoords;
};

inline void Earth::initialize(Engine *e, int sl, int st, float sp) {
    engine = e;
    slices = sl;
    stacks = st;
    spherical = sp;
    tarSpherical = sp;
    texture = engine->loadTexture(Config::textureFile);

    float xDif = 360/(slices);
    float yDif = 180 /(stacks);
    float x, y;
    nVertices = (stacks + 1)*(slices + 1);
    nTriangles = 2*slices*stacks;
    verts = new vec3[nVertices];
    faces = new int[3*nTriangles];
    texCoords = new vec2[nVertices];
    normals = new vec3[nVertices];
    int i = sizeof(faces);
    y = 90;
    for (int i = 0; i < stacks+1; i++) {
        x = -180;
        for (int j = 0; j < slices+1; j++) {
            verts[i*(slices + 1) + j] = getPosition(y,x);
            texCoords[i*(slices + 1) + j] = vec2((x+180.0)/360.0, (y-90)/-180);
            normals[i*(slices + 1) + j] = getNormal(y,x);
            x += xDif;
        }
        y -= yDif;
    }

    //assign upper right, then lower left of quad triangles
    for(int i = 0; i < stacks; i++) {
        for(int j = 0; j < slices; j++) {
            //upper right
            faces[(i*6*slices) + j*6] = i*(slices+1) + j;
            faces[(i*6*slices) + j*6 + 1] = i*(slices + 1) + j + 1;
            faces[(i*6*slices) + j*6 + 2] = (i+1)*(slices+1) + j + 1;

            faces[(i * 6 * slices) + j * 6 + 3] = i*(slices + 1) + j;
            faces[(i * 6 * slices) + j * 6 + 4] = (i+1)*(slices + 1) + j + 1;
            faces[(i * 6 * slices) + j * 6 + 5] = (i + 1)*(slices + 1) + j ;
        }
    }
    normalBuffer = engine->allocateVertexBuffer(nVertices*sizeof(vec3));
    engine->copyVertexData(normalBuffer, normals, nVertices*sizeof(vec3));
    texCoordBuffer = engine->allocateVertexBuffer(nVertices*sizeof(vec2));
    engine->copyVertexData(texCoordBuffer, texCoords, nVertices*sizeof(vec2));
    vertBuffer = engine->allocateVertexBuffer(nVertices*sizeof(vec3));
    engine->copyVertexData(vertBuffer, verts, nVertices*sizeof(vec3));
    faceBuffer = engine->allocateElementBuffer(nTriangles*3*sizeof(int));
    engine->copyElementData(faceBuffer, faces, nTriangles*3*sizeof(int));
}

inline float Earth::isSpherical() {
    return spherical;
}

inline void Earth::setSpherical(float s) {
    spherical = s;
    float xDif = 360 / (slices);
    float yDif = 180 / (stacks);
    float x, y;
    y = 90;
    for (int i = 0; i < stacks+1; i++) {
        x = -180;
        for (int j = 0; j < slices+1; j++) {
            verts[i*(slices + 1) + j] = getPosition(y,x);
            normals[i*(slices + 1) + j] = getNormal(y,x);
            x += xDif;
        }
        y -= yDif;
    }
    engine->copyVertexData(normalBuffer, normals, nVertices*sizeof(vec3));
    engine->copyVertexData(vertBuffer, verts, nVertices*sizeof(vec3));
}

inline float Earth::getTargetSpherical() {
    return tarSpherical;
}

inline void Earth::setTargetSpherical(float s) {
    tarSpherical = s;
}

inline vec3 Earth::getPosition(float latitude, float longitude) {
    // pi/180 = 0.01745329251
    if (spherical == 0){
        float rectLat = latitude * 0.01745329251;
        float rectLong = longitude * 0.01745329251;
        return vec3(rectLong,rectLat, 0);
    }
    else if (spherical == 1){
        float phi = latitude * 0.01745329251;
        float theta = longitude * 0.01745329251;
        return vec3(cos(phi)*sin(theta), sin(phi), cos(theta)*cos(phi));
    }
        
    else {
        latitude  = latitude * 0.01745329251;
        longitude =  longitude * 0.01745329251;
        vec3 rectCoords = vec3(longitude,latitude, 0);
        vec3 sphericalCoords = vec3(cos(latitude)*sin(longitude), sin(latitude), cos(longitude)*cos(latitude));
        return lerp(rectCoords, sphericalCoords, spherical);
    }
}

inline vec3 Earth::getNormal(float latitude, float longitude) {
    if (spherical == 0){
        return vec3(0,0,1);
    }
    else if (spherical == 1){
        latitude  = latitude * 0.01745329251;
        longitude =  longitude * 0.01745329251;
        return vec3(cos(latitude)*sin(longitude), sin(latitude), cos(longitude)*cos(latitude));
    }
    else {
        latitude  = latitude * 0.01745329251;
        longitude =  longitude * 0.01745329251;
        vec3 sphericalCoords = vec3(cos(latitude)*sin(longitude), sin(latitude), cos(longitude)*cos(latitude));
        return lerp(vec3(0,0,1), sphericalCoords, spherical);
    }
}

inline void Earth::draw(bool textured) {
    if(textured) {
        engine->setTexCoordArray(texCoordBuffer);
        engine->setTexture(texture);
    }
    engine->setVertexArray(vertBuffer);
    engine->setNormalArray(normalBuffer);
    engine->drawElements(GL_TRIANGLES, faceBuffer, nTriangles*3);
    engine->unsetTexture();
}

#endif
